"""
pylhc-submitter
~~~~~~~~~~~~~~~~

pylhc-submitter contains scripts to simplify the creation and submission of jobs to HTCondor at CERN.

:copyright: pyLHC/OMC-Team working group.
:license: MIT, see the LICENSE.md file for details.
"""
__title__ = "pylhc_submitter"
__description__ = "pylhc-submitter contains scripts to simplify the creation and submission of jobs to HTCondor at CERN"
__url__ = "https://github.com/pylhc/submitter"
__version__ = "2.0.2"
__author__ = "pylhc"
__author_email__ = "pylhc@github.com"
__license__ = "MIT"

__all__ = [__version__]
